"""Configuration of the timings."""


class TimingConfig:  # pylint: disable = too-few-public-methods
    """Global configuration of timing."""

    enable_cache: bool = True
    overhead: float = 0.0



"""Utility and supporting functions for timing module."""

import collections
import logging
import statistics
import typing as t

from .config import TimingConfig
from .timing import Timing
from .group import TimingGroup
from .cache import TimingCache

if __debug__:
    _LOG = logging.getLogger(__name__)


def get_timing_group(*name_fragments: str) -> TimingGroup:
    """Work similarly to logging.getLogger()."""
    assert name_fragments
    assert all(isinstance(_, str) and _ for _ in name_fragments), name_fragments
    name = '.'.join(name_fragments)
    if name in TimingCache.flat:
        return TimingCache.flat[name]

    if TimingConfig.enable_cache:
        timing_cache = TimingCache.hierarchical

        name_fragments_normalized = name.split('.')
        for i, name_fragment in enumerate(name_fragments_normalized):
            if name_fragment not in timing_cache:
                for level in range(i, len(name_fragments_normalized)):
                    timing_cache[name_fragments_normalized[level]] = collections.OrderedDict()
                    timing_cache = timing_cache[name_fragments_normalized[level]]
                break
            timing_cache = timing_cache[name_fragment]

    timing_group = TimingGroup(name)

    if TimingConfig.enable_cache:
        timing_cache['.'] = timing_group
        TimingCache.flat[name] = timing_group

    return timing_group


def query_cache(*name_fragments: str) -> t.Union[dict, TimingGroup, Timing]:
    """Request timing data from global cache."""
    return TimingCache.query(*name_fragments)


def normalize_overhead(samples: int = 10000, threshold: float = 1.0):
    """Investigate overhead of starting and stopping the timer.

    Do it so as to take the overhead into account when calculating actual execution times.
    """
    assert isinstance(samples, int)
    assert isinstance(threshold, float)

    timing_overhead = Timing('timing overhead test')
    overheads = []

    cache_status = TimingConfig.enable_cache
    TimingConfig.enable_cache = False
    __ = TimingGroup('timing_overhead_normalization')
    for _ in __.measure_many('overhead', samples=samples, threshold=threshold):
        timing_overhead.start()
        timing_overhead.stop()
        overheads.append(timing_overhead.elapsed)

    TimingConfig.enable_cache = cache_status

    mean: float = statistics.mean(overheads)  # type: ignore
    median: float = statistics.median(overheads)  # type: ignore
    stdev: float = statistics.pstdev(overheads, mean)  # type: ignore
    variance: float = statistics.pvariance(overheads, mean)  # type: ignore

    TimingConfig.overhead = median

    if __debug__:
        _LOG.log(
            logging.DEBUG if all(_ <= 0.00001 for _ in (mean, median, stdev, variance))
            else logging.WARNING,
            'normalized overhead using %i samples: mean=%f, median=%f, stdev=%f, variance=%f',
            samples, mean, median, stdev, variance)

        if stdev > 0.0001 or variance > 0.0001:
            _LOG.error(
                'stdev=%f and/or variance=%f are large -- timing results will not be stable',
                stdev, variance)

        if mean > 0.0001 or median > 0.0001:
            _LOG.error(
                'mean=%f and/or median=%f are large -- timing will incur overhead',
                mean, median)


normalize_overhead()



"""Tests of utility functions."""

import logging
from time import perf_counter as original_perf_counter
import time
import unittest
import unittest.mock

import numpy as np

from timing.config import TimingConfig
from timing.cache import TimingCache
from timing.utils import get_timing_group, query_cache, normalize_overhead


def slow_perf_counter():
    time.sleep(0.1)
    return original_perf_counter()


def erratic_perf_counter():
    time.sleep(np.random.uniform(0.0, 0.1))
    return original_perf_counter()


class Tests(unittest.TestCase):

    @unittest.skipIf(not __debug__, reason='skipping test which would fail for optimised run')
    def test_get_timing_group(self):
        self.assertTrue(TimingConfig.enable_cache)
        TimingConfig.enable_cache = False

        timers = get_timing_group('timings.getting_group')
        self.assertIsNot(timers, get_timing_group('timings.getting_group'))

        TimingConfig.enable_cache = True

        timers = get_timing_group('timings.getting_group')
        self.assertIs(timers, get_timing_group('timings.getting_group'))
        self.assertIs(timers, get_timing_group('timings', 'getting_group'))
        with self.assertRaises(AssertionError):
            get_timing_group(32)

    def test_query_cache(self):
        with self.assertRaises(KeyError):
            query_cache('timings.non_existing_group')
        timers = get_timing_group('timings.existing_group')
        self.assertIs(timers, query_cache('timings.existing_group'))
        self.assertIs(timers, query_cache('timings', 'existing_group'))

    @unittest.skipIf(not __debug__, reason='skipping test which would fail for optimised run')
    def test_overhead(self):
        self.assertTrue(TimingConfig.enable_cache)
        TimingCache.clear()

        normalize_overhead()
        self.assertLessEqual(TimingConfig.overhead, 0.0001)

        with self.assertLogs(level=logging.ERROR) as log:
            with unittest.mock.patch.object(time, 'perf_counter', new=slow_perf_counter):
                normalize_overhead()
        self.assertGreaterEqual(TimingConfig.overhead, 0.1)
        self.assertTrue(any(all(_ in line for _ in ('ERROR', 'mean', 'median', 'large'))
                            for line in log.output), msg=log.output)

        with self.assertLogs(level=logging.ERROR) as log:
            with unittest.mock.patch.object(time, 'perf_counter', new=erratic_perf_counter):
                normalize_overhead()
        self.assertTrue(any(all(_ in line for _ in ('ERROR', 'variance', 'stdev', 'large'))
                            for line in log.output), msg=log.output)
        self.assertTrue(any(all(_ in line for _ in ('ERROR', 'mean', 'median', 'large'))
                            for line in log.output), msg=log.output)

        normalize_overhead()
        self.assertLessEqual(TimingConfig.overhead, 0.0001)

        self.assertTrue(not TimingCache.hierarchical, msg=TimingCache.hierarchical)
        self.assertTrue(not TimingCache.flat, msg=TimingCache.flat)
        self.assertTrue(not TimingCache.chronological, msg=TimingCache.chronological)



"""Initialization of tests for timing package."""

import boilerplates.logging


class TestsLogging(boilerplates.logging.Logging):
    """Test logging configuration."""

    packages = ['timing']


TestsLogging.configure()



"""Tests for the time measurement itself."""

import copy
import logging
import time
import unittest

from timing.timing import Timing, TimingState

_LOG = logging.getLogger(__name__)


class Tests(unittest.TestCase):

    def test_name(self):
        timer = Timing('timing')
        self.assertEqual(timer.name, 'timing')
        self.assertEqual(timer.state, TimingState.NOT_STARTED)
        with self.assertRaises(AssertionError):
            _LOG.debug('%f', timer.begin)
        with self.assertRaises(AssertionError):
            _LOG.debug('%f', timer.end)
        with self.assertRaises(AssertionError):
            _LOG.debug('%f', timer.elapsed)

    def test_start(self):
        timer = Timing('timing')
        timer.start()
        self.assertEqual(timer.state, TimingState.RUNNING)
        time.sleep(0.01)
        self.assertIsInstance(timer.begin, float)
        self.assertGreater(timer.begin, 0)
        with self.assertRaises(AssertionError):
            _LOG.debug('%f', timer.end)
        with self.assertRaises(AssertionError):
            _LOG.debug('%f', timer.elapsed)

    def test_stop(self):
        timer = Timing('timing')
        timer.start()
        time.sleep(0.01)
        timer.stop()
        self.assertEqual(timer.state, TimingState.FINISHED)
        self.assertIsInstance(timer.end, float)
        self.assertGreater(timer.end, 0)
        self.assertIsInstance(timer.elapsed, float)
        self.assertGreater(timer.elapsed, 0)

    def test_stop_not_running(self):
        timer = Timing('timing')
        with self.assertRaises(AssertionError):
            timer.stop()
        with self.assertRaises(AssertionError):
            _LOG.debug('%f', timer.end)

    def test_compare(self):
        timer = Timing('timing')
        other_timer = Timing('timing')
        self.assertEqual(timer, other_timer)
        # start
        timer.start()
        self.assertNotEqual(timer, other_timer)
        other_timer.start()
        self.assertEqual(timer.state, other_timer.state)
        self.assertNotEqual(timer, other_timer)
        other_timer = copy.deepcopy(timer)
        time.sleep(0.01)
        self.assertEqual(timer, other_timer)
        # stop
        timer.stop()
        self.assertNotEqual(timer, other_timer)
        other_timer.stop()
        self.assertEqual(timer.state, other_timer.state)
        self.assertNotEqual(timer, other_timer)
        other_timer = copy.deepcopy(timer)
        self.assertEqual(timer, other_timer)

    def test_output(self):
        timer = Timing('my_timing')
        self.assertIn('my_timing', str(timer))
        self.assertIn('my_timing', repr(timer))



"""Core module of the timing package."""

import enum
import time
import typing as t


@enum.unique
class TimingState(enum.IntEnum):
    """State of a timing."""

    NOT_STARTED = 0
    RUNNING = 1
    FINISHED = 2


class Timing:
    """Timing of performance-critical parts of application.

    Uses time.perf_counter(): https://docs.python.org/3/library/time.html#time.perf_counter
    """

    def __init__(self, name: str):
        assert isinstance(name, str), type(name)
        assert name
        # self._group = None  # type: t.Optional[TimingGroup]
        self._name: str = name
        self._state: int = 0
        self._begin: t.Optional[float] = None
        self._end: t.Optional[float] = None
        self._elapsed: t.Optional[float] = None

    def _calculate_elapsed(self) -> None:
        assert self._begin is not None, 'timing has not started yet'
        assert self._end is not None, 'timing has not finished yet'
        self._elapsed = self._end - self._begin

    @property
    def name(self) -> str:
        return self._name

    @property
    def begin(self) -> float:
        assert self._begin is not None, 'timing has not started yet'
        return self._begin

    @property
    def end(self) -> float:
        assert self._end is not None, 'timing has not finished yet'
        return self._end

    @property
    def elapsed(self) -> float:
        assert self._elapsed is not None, 'timing has not finished yet'
        return self._elapsed

    @property
    def state(self) -> TimingState:
        return {
            0: TimingState.NOT_STARTED,
            1: TimingState.RUNNING,
            2: TimingState.FINISHED,
        }[self._state]

    def start(self) -> None:
        """Start the timer."""
        self._state = 1
        self._end = None
        self._elapsed = None
        self._begin = time.perf_counter()

    def stop(self) -> None:
        """Stop the timer."""
        self._end = time.perf_counter()
        if self.state is not TimingState.RUNNING:
            self._end = None
        assert self.state is TimingState.RUNNING, 'timing has not started yet'
        self._state = 2
        self._calculate_elapsed()

    def __eq__(self, other):
        if not isinstance(other, Timing) or self.state is not other.state \
                or self._name != other.name:
            return False
        if self._state == 0:
            return True
        if self._begin != other.begin:
            return False
        if self.state == 1:
            return True
        return self._end == other.end

    def __str__(self):
        args = [self._name, self._begin, self._end, self._elapsed]
        return f'{type(self).__name__}({", ".join([str(_) for _ in args])})'

    def __repr__(self):
        return str(self)



"""Initialization of timing package."""

__all__ = [
    'TimingConfig', 'Timing', 'TimingGroup', 'TimingCache', 'get_timing_group', 'query_cache']

from .config import TimingConfig
from .timing import Timing
from .group import TimingGroup
from .cache import TimingCache
from .utils import get_timing_group, query_cache



"""Setup script for timing package."""

import boilerplates.setup


class Package(boilerplates.setup.Package):
    """Package metadata."""

    name = 'timing'
    description = 'Simplify logging of timings of selected parts of an application.'
    url = 'https://github.com/mbdevpl/timing'
    classifiers = [
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Scientific/Engineering',
        'Topic :: System :: Benchmark',
        'Topic :: System :: Logging',
        'Typing :: Typed']
    keywords = ['timing', 'timer', 'time measurement', 'profiling', 'reproducibility']


if __name__ == '__main__':
    Package.setup()



"""Handling of group of timings."""

import contextlib
import datetime
import functools
import types
import typing as t

import numpy as np

from .config import TimingConfig
from .timing import Timing


class TimingGroup(dict):
    """Group of timings."""

    def __init__(self, name: str):
        super().__init__()
        assert isinstance(name, str)

        self._name: str = name
        self._timings: t.List[Timing] = []
        self._summary: t.Optional[t.Dict[str, t.Any]] = None

    @property
    def name(self) -> str:
        return self._name

    @property
    def timings(self) -> t.List[Timing]:
        return list(self._timings)

    @property
    def summary(self) -> t.Dict[str, t.Any]:
        """Return a collection of statistics for the timings in this group.

        Calculate statistics if not already calculated, and use cached values otherwise.
        """
        if self._summary is None:
            self.summarize()
        assert self._summary is not None
        return self._summary

    def start(self, name: str) -> Timing:
        """Create a Timing belonging to this TimingGroup and start it."""
        if '.' in name:
            from .utils import get_timing_group  # pylint: disable = import-outside-toplevel
            prefix, _, suffix = name.rpartition('.')
            group = get_timing_group(self._name, prefix)
            return group.start(suffix)

        timing = Timing(name)
        if TimingConfig.enable_cache:
            from .cache import TimingCache  # pylint: disable = import-outside-toplevel
            self._timings.append(timing)
            if self._name in TimingCache.flat and TimingCache.flat[self._name] is self:
                cache_entry = (datetime.datetime.now(), timing)
                TimingCache.chronological.append(cache_entry)
        if name in self:
            self[name].append(timing)
        else:
            self[name] = [timing]
        timing.start()
        return timing

    def measure(self, function_or_name: t.Callable | str | None = None, name: str | None = None):
        """Use this method as a context manager or decorator.

        As context manager:

        with ....measure(name) as timer:
            ...

        As decorator:

        @measure
        def ...

        @measure(name)
        def ...
        """
        if function_or_name is not None:
            if isinstance(function_or_name, str):
                assert name is None, 'name given in the first argument, 2nd argument must be None'
                function, name = None, function_or_name
            else:
                function = function_or_name
        else:
            assert name is not None, 'at least one argument to measure() is required'
            function = None
        if function is None:
            # in practice this path is also taken when @measure(name) is used,
            # but since contextlib uses ContextDecorator, it works
            assert name is not None
            return self._measure_context(name)
        assert isinstance(function, types.FunctionType)
        return self._measure_decorator(function, name)

    @contextlib.contextmanager
    def _measure_context(self, name: str) -> t.Generator[Timing, None, None]:
        """Return the just-started timer as context variable."""
        timer = self.start(name)
        yield timer
        timer.stop()

    def _measure_decorator(self, function: types.FunctionType, name: str | None = None):
        """Return the original function wrapped in a timing context."""
        if name is None:
            name = function.__name__

        @functools.wraps(function)
        def function_wrapper(*args, **kwargs):
            with self._measure_context(name):
                return function(*args, **kwargs)
        return function_wrapper

    def measure_many(self, name: str, samples: t.Optional[int] = None,
                     threshold: t.Optional[float] = None) -> t.Iterator[Timing]:
        """Iterate and time each iteration until some iterations or until some time passes.

        Use via 'for timer in measure_many('name'[, samples][, threshold]).
        """
        assert samples is not None or threshold is not None, (samples, threshold)
        assert samples is None or isinstance(samples, int) and samples > 0, samples
        assert threshold is None or threshold > 0, threshold
        while True:
            timer = self.start(name)
            yield timer
            timer.stop()
            if samples is not None:
                samples -= 1
                if samples == 0:
                    break
            if threshold is not None:
                assert timer.elapsed is not None
                threshold -= timer.elapsed
                if threshold <= 0:
                    break

    def query_cache(self, *name_fragments: str) -> t.Union[dict, 'TimingGroup', Timing]:
        """Query the cache within the scope of this timing group."""
        from .cache import TimingCache  # pylint: disable = import-outside-toplevel
        if not name_fragments:
            return TimingCache.query(self._name)
        return TimingCache.query(self._name, *name_fragments)

    def summarize(self) -> None:
        """Calculate (or recalculate) various statistics from the raw data."""
        self._summary = {}
        for name, timings in self.items():
            elapsed = [_.elapsed for _ in timings]
            # np.ndarray((len(timings),), dtype=float)
            array = np.array(elapsed, dtype=float)
            self._summary[name] = {
                'data': array.tolist(),
                'samples': len(array),
                'min': array.min(),
                'max': array.max(),
                'mean': array.mean(),
                'median': np.median(array),
                'var': array.var(),
                'stddev': array.std()}

    def __eq__(self, other):
        if not isinstance(other, TimingGroup):
            return False
        return self._name == other.name and self._timings == other.timings

    def __str__(self):
        args = [self._name] + self._timings
        return f'{type(self).__name__}({", ".join([str(_) for _ in args])})'

    def __repr__(self):
        return str(self)



"""Cache of timing results."""

import collections
import datetime
import typing as t

from .timing import Timing
from .group import TimingGroup


class TimingCache:
    """Global cache for timing results."""

    hierarchical: t.Dict[str, dict] = collections.OrderedDict()
    """Hierarchy of TimingGroup objects structured in a tree that branches
    at '.' in the TimingGroup names.

    For example, having timing groups named after some example modules:

    - TimingGroup('spam.eggs.spam')
    - TimingGroup('spam.spam')
    - TimingGroup('eggs.ham')
    - TimingGroup('spam.eggs')

    the hierarchy would be:

    hierarchical_timing_cache = {
        'spam': {
            'eggs': {
                '.': TimingGroup('spam.eggs'),
                'spam': {
                    '.': TimingGroup('spam.eggs.spam')
                    }
                },
            'spam': {
                '.': TimingGroup('spam.spam')
                }
            },
        'eggs': {
            'ham': {
                '.': TimingGroup('eggs.ham')
                }
            }
        }
    """

    flat: t.Dict[str, TimingGroup] = collections.OrderedDict()
    """Flattened view of the TimingGroup objects hierarchy.

    For example, having timing groups named after some example modules:

    - TimingGroup('spam.eggs.spam')
    - TimingGroup('spam.spam')
    - TimingGroup('eggs.ham')
    - TimingGroup('spam.eggs')

    the flattened view would simply be:

    flat_timing_cache = {
        'spam.eggs.spam': TimingGroup('spam.eggs.spam')
        'spam.spam': TimingGroup('spam.spam')
        'eggs.ham': TimingGroup('eggs.ham')
        'spam.eggs': TimingGroup('spam.eggs')
        }
    """

    chronological: t.List[t.Tuple[datetime.datetime, Timing]] = []
    """Individual Timing objects in the order that they were started in."""

    @classmethod
    def clear(cls) -> None:
        cls.hierarchical = collections.OrderedDict()
        cls.flat = collections.OrderedDict()
        cls.chronological = []

    @classmethod
    def query(cls, *name_fragments: str) -> t.Union[dict, TimingGroup, Timing]:
        """Query the cache using one or more name fragments."""
        assert name_fragments
        assert all(isinstance(_, str) and _ for _ in name_fragments), name_fragments
        normalized_name_fragments = '.'.join(name_fragments).split('.')
        timing_cache = TimingCache.hierarchical
        for _, name_fragment in enumerate(normalized_name_fragments):
            timing_cache = timing_cache[name_fragment]
        return timing_cache['.']



"""Tests of handling of group of timings."""

import contextlib
import logging
import time
import types
import unittest

from timing.group import TimingGroup
from timing.utils import get_timing_group, query_cache

_LOG = logging.getLogger(__name__)


class Tests(unittest.TestCase):

    def test_basic(self):
        timers = TimingGroup('timings.basic')
        self.assertFalse(timers == 'timings.basic')
        other_timers = TimingGroup('timings.basic')
        self.assertEqual(timers, other_timers)
        self.assertEqual(timers.name, 'timings.basic')
        self.assertListEqual(timers.timings, [])
        self.assertEqual(timers.summary, {})
        timer = timers.start('timer')
        time.sleep(0.01)
        timer.stop()
        timer = timers.start('sub.timer')
        time.sleep(0.01)
        timer.stop()
        _LOG.info('%s', timers)
        _LOG.info('%r', timers)
        self.assertEqual(timers, timers)
        self.assertNotEqual(timers, other_timers)

    def test_measure_context(self):
        timers = TimingGroup('timings.contexts')

        with timers.measure('context1'):
            time.sleep(0.01)

        self.assertIn('context1', timers.summary)
        self.assertEqual(timers.summary['context1']['samples'], 1)

    def test_measure_decorator(self):
        timers = TimingGroup('timings.decorators')

        @timers.measure
        def add(num1, num2):
            time.sleep(0.02)
            return num1 + num2

        self.assertNotIsInstance(add, contextlib.ContextDecorator)
        self.assertIsInstance(add, types.FunctionType)

        @timers.measure('mult')
        def multiply(num1, num2):
            time.sleep(0.02)
            return num1 * num2

        self.assertNotIsInstance(multiply, contextlib.ContextDecorator)
        self.assertIsInstance(multiply, types.FunctionType)

        self.assertEqual(add(21, 21), 42)
        self.assertEqual(add(-5, 5), 0)
        self.assertEqual(multiply(21, 2), 42)
        self.assertEqual(multiply(-5, 5), -25)
        self.assertIn('add', timers.summary)
        self.assertEqual(timers.summary['add']['samples'], 2)
        self.assertNotIn('multiply', timers.summary)
        self.assertIn('mult', timers.summary)
        self.assertEqual(timers.summary['mult']['samples'], 2)

    def test_measure_decorator_name_only(self):
        timers = TimingGroup('timings.decorators')

        @timers.measure(name='mult')
        def multiply(num1, num2):
            time.sleep(0.02)
            return num1 * num2

        self.assertEqual(multiply(21, 2), 42)
        _LOG.info('%s', timers)

    def test_context_or_decorator(self):
        timers = TimingGroup('timings.context_or_decorator')

        def sub(num1, num2):
            time.sleep(0.001)
            return num1 - num2

        context = timers.measure('sub')
        self.assertIsInstance(context, contextlib.ContextDecorator)
        self.assertNotIsInstance(context, types.FunctionType)
        decorated = timers.measure(sub)
        self.assertNotIsInstance(decorated, contextlib.ContextDecorator)
        self.assertIsInstance(decorated, types.FunctionType)
        self.assertEqual(decorated(10, 11), -1)
        named_decorated = timers.measure(sub, 'sub')
        self.assertNotIsInstance(named_decorated, contextlib.ContextDecorator)
        self.assertIsInstance(named_decorated, types.FunctionType)
        self.assertEqual(named_decorated(100, 110), -10)

    def test_measure_many(self):
        timers = TimingGroup('timings.many')
        for _ in timers.measure_many('by_samples', samples=10):
            time.sleep(0.001)
        for _ in timers.measure_many('by_threshold', threshold=0.01):
            time.sleep(0.001)
        for _ in timers.measure_many('by_both', samples=10, threshold=0.01):
            time.sleep(0.001)
        self.assertGreaterEqual(len(timers.timings), 12)

    def test_query_cache(self):
        timers = get_timing_group('timings.root_group.subgroup')
        with timers.measure('context1'):
            time.sleep(0.01)

        root_timers = get_timing_group('timings.root_group')
        self.assertIs(timers, timers.query_cache())
        self.assertIs(timers, root_timers.query_cache('subgroup'))
        self.assertIs(timers, query_cache('timings.root_group.subgroup'))
        self.assertIs(timers, query_cache('timings.root_group', 'subgroup'))
        self.assertIs(timers, query_cache('timings', 'root_group', 'subgroup'))
